/* 
 * Copyright (c) 2005-2006 Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#define GL_GLEXT_PROTOTYPES

#include <cmath>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "texture.h"
#include "heightfield.h"


HeightField::HeightField (std::string heightmap,
                          std::string lightmap,
                          std::string detailmap)
{
	data = 0;
	lighttex  = new Texture (lightmap.c_str(), false, true);
	detailtex = new Texture (detailmap.c_str(), false, true);

	surface = IMG_Load (heightmap.c_str());
	if (!surface) {
		fprintf (stderr,"Unable to load map %s", heightmap.c_str());
		return;
	}
	if ((surface->format->BitsPerPixel != 8) || (surface->format->BytesPerPixel != 1)) {
		fprintf (stderr,"Map image is not a greyscale one");
		return;
	}
	if (surface->w != surface->h) {
		fprintf (stderr,"Map image is not square");
		return;
	}

	data = (unsigned char *)(surface->pixels);
}

HeightField::~HeightField()
{}


void HeightField::make (int size_)
{
	// No map loaded => return
	if (!data)
		return;

	// size is the number of vertex in a row in the resulting mesh
	// for example, we can have a 1024x1024 map but we want a 64x64 mesh
	// so we need to subsample the map
	size = size_;
	int step;
	step = surface->w / size;
	if (step < 1) {
		size = surface->w;
		step = 1;
	}

	float lighttex_delta = 1.0f/surface->w;
	float detailtex_delta = 32.0f/surface->w;
	Vertex v;
	int i = 0;

	// Loop through pixels in the map
	n_vertices = size*size;
	vertices   = new Vertex[n_vertices];

	for (int y=0; y < size; y++) {
		for (int x=0; x < size; x++) {
			i = y*size + x;
			v.x = ((float)x/float(size) - 0.5f);
			v.z = ((float)y/float(size) - 0.5f);
			v.y = (float)(data[y*step*surface->w + x*step])/256.0f;
			v.u0 = (float)x*step        * detailtex_delta;
			v.v0 = 1.0f - (float)y*step * detailtex_delta;
			v.u1 = (float)x*step        * lighttex_delta;
			v.v1 = 1.0f - (float)y*step * lighttex_delta;
			vertices[i] = v;
		}
	}

	// Make triangles indices
	n_indices = (size_-1)*(size_-1)*6;
	indices = new GLuint [n_indices];
	int j = 0;
	for (int y=0; y < size-1; y++) {
		for (int x=0; x < size-1; x++) {
			i = y*(size)+x;
			indices[j++] = i;
			indices[j++] = i+1;
			indices[j++] = i+size;
			indices[j++] = i+size+1;
			indices[j++] = i+size;
			indices[j++] = i+1;
		}
	}
}


void
HeightField::render (void) 
{
    static GLuint list = 0;
    if (list) {
        glCallList (list);
        return;
    }

    list = glGenLists (1);
    glNewList (list, GL_COMPILE_AND_EXECUTE);

    glPolygonOffset (1.0f, 1.0f);
    glEnable (GL_POLYGON_OFFSET_FILL);
    glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
    glEnable (GL_TEXTURE_2D);
    glColor3f (1,1,1);

	glVertexPointer (3, GL_FLOAT, sizeof(Vertex), &vertices[0].x);

    // Texture Unit 0
    glActiveTextureARB (GL_TEXTURE0_ARB);
    glClientActiveTextureARB (GL_TEXTURE0_ARB);
    glEnable (GL_TEXTURE_2D);

    glTexCoordPointer (2, GL_FLOAT, sizeof(Vertex), &vertices[0].u0);
    glBindTexture (GL_TEXTURE_2D, detailtex->id);
    glEnableClientState (GL_TEXTURE_COORD_ARRAY);
    glEnableClientState (GL_VERTEX_ARRAY);


	// Texture Unit 1
	glActiveTextureARB (GL_TEXTURE1_ARB);
    glClientActiveTextureARB (GL_TEXTURE1_ARB);
	glEnable (GL_TEXTURE_2D);

	// Set up the blending of the 2 textures
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glTexCoordPointer (2, GL_FLOAT, sizeof(Vertex), &vertices[0].u1);
	glBindTexture (GL_TEXTURE_2D, lighttex->id);
    glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glEnableClientState (GL_VERTEX_ARRAY);


	// Render the Terrain
	glDrawElements(GL_TRIANGLES, n_indices, GL_UNSIGNED_INT, indices);


	// Switch off Texture Unit 1
	glDisable (GL_TEXTURE_2D);
	glDisableClientState (GL_VERTEX_ARRAY);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);

	// Switch off of Texture Unit 0
	glActiveTextureARB (GL_TEXTURE0_ARB);
	glClientActiveTextureARB (GL_TEXTURE0_ARB);
    glDisable (GL_TEXTURE_2D);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	glDisableClientState (GL_VERTEX_ARRAY);


    /*
      Wireframe

      glColor3f (1,1,1);
      glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
      glDisable (GL_POLYGON_OFFSET_FILL);
      glVertexPointer (3, GL_FLOAT, sizeof(Vertex), &vertices[0].x);
      glEnableClientState (GL_VERTEX_ARRAY);
      glDrawElements(GL_TRIANGLES, n_indices, GL_UNSIGNED_INT, indices);
      glDisableClientState (GL_VERTEX_ARRAY);
    */
    glEndList();
}

