/* 
 * Copyright (c) 2005-2006 Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <iostream>
#include <cstdlib>
#include "../config.h"
#include "scene.h"
#include "canvas.h"
#include "canvas_sdl.h"
#include "canvas_glx.h"

#ifdef WIN32
#  include <windows.h>
#else
#  include <argp.h>
#endif


Canvas *canvas;
Scene   scene;

bool	 use_glx	= false;
long int window_id  = 0;
int      fps        = 100;
bool	 fullscreen = false;
bool     panoramic  = false;

#ifdef WIN32

#define OPTION_HIDDEN 1
#define OPTION_DOC 2

#define ARGP_KEY_ARG 0
#define ARGP_ERR_UNKNOWN 0

#define ARGP_LONG_ONLY 1

typedef int (*argp_parser_t)(int, char*, struct argp_state*);

struct argp_option {
    const char *name;
    int key;
    const char *arg;
    int flags;
    const char *doc;
    int group;
};

struct argp {
    const struct argp_option *options;
    argp_parser_t parser;
    const char *args_doc;
    const char *doc;
    // there are more members, but this is all I need
};

struct argp_state {
    int argc;
    char **argv;
    int next;
    char *name;
    // there are more members, but this is all I need
};

void argp_help(struct argp *argp_s)
{
    std::cerr << argp_s->doc << std::endl;

    for (int i = 0; argp_s->options[i].name; i++) {
	if (argp_s->options[i].flags & OPTION_HIDDEN)
	    continue;
	if (argp_s->options[i].flags & OPTION_DOC)
	    std::cerr << argp_s->options[i].name << std::endl;
	else {
	    if (isprint(argp_s->options[i].key))
		std::cerr << "  -" << (char)argp_s->options[i].key << ",";
	    else
		std::cerr << "     ";
	    std::cerr << " --" << argp_s->options[i].name << "\t";
	    std::cerr << argp_s->options[i].doc << std::endl;
	}
    }
}

// doesn't work exactly as real argp, but suits this program
int argp_parse(struct argp *argp_s, int argc, char *argv[], int, int *, void *)
{
    char *opt, *arg=0;
    int optind = 0;

    struct argp_state state;
    state.argc = argc;
    state.argv = argv;
    state.name = argv[0];

    for (int i = 1; i < state.argc; i = state.next) {
	opt = state.argv[i];
	arg = 0;
	optind = -1;
	state.next = i+1;

	if (opt[0] != '-') {
	    if ((*argp_s->parser)(ARGP_KEY_ARG, 0, &state) < 0)
		return -1;
	    continue;
	}
	opt++;
	if (*opt == '-')
	    opt++;

	if ((arg = strchr(opt, '=')))
	    *arg++ = 0;

	if (opt[1] == 0) { // short option
	    for (int j = 0; argp_s->options[j].name; j++) {
		if (opt[0] == argp_s->options[j].key) {
		    optind = j;
		    break;
		}
	    }
	    if (optind >= 0)
		;
	    else if (opt[0] == 'h') {
		argp_help(argp_s);
		return -1;
	    }
	    else if (opt[0] == 'u') {
		argp_help(argp_s);
		return -1;
	    }
	    else if (opt[0] == 'V') {
		extern char *argp_program_version;
		std::cerr << argp_program_version << std::endl;
		return -1;
	    }
	}
	else { // long option
	    for (int j = 0; argp_s->options[j].name; j++) {
		if (strcmp(opt, argp_s->options[j].name)==0) {
		    optind = j;
		    break;
		}
	    }
	    if (optind >= 0)
		;
	    else if (strcmp(opt, "help") == 0) {
		argp_help(argp_s);
		return -1;
	    }
	    else if (strcmp(opt, "usage") == 0) {
		argp_help(argp_s);
		return -1;
	    }
	    else if (strcmp(opt, "version") == 0) {
		extern char *argp_program_version;
		std::cerr << argp_program_version << std::endl;
		return -1;
	    }
	}
	if (optind < 0) { // doesn't exist
	    std::cerr << state.name << ": unrecognized option `"
		 << state.argv[i] << "'" << std::endl;
	    std::cerr << "Try `" << state.name << " --help' or `"
		 << state.name << " --usage' for more information" << std::endl;
	    return -1;
	}
	if (argp_s->options[optind].arg != 0 && arg == 0) {
	    state.next = i+2;
	    if (i+1 >= state.argc) {
		std::cerr << state.name << ": option requires an argument -- "
		    << opt << std::endl;
		return -1;
	    }
	    arg = state.argv[i+1];
	}
	if ((*argp_s->parser)(argp_s->options[optind].key, arg, &state) < 0)
	    return -1;
    }
    return 0;
}

#endif


#define OPT_FULLSCREEN	1
#define OPT_FPS 		2

const char *argp_program_version =
"Hills " PACKAGE_VERSION " by Nicolas Rougier <Nicolas.Rougier@loria.fr>";

static char doc[] =
"Peaceful fly over grass hills.";

static struct argp_option options[] = {
{"window-id",	'W', 			"WinID", OPTION_HIDDEN},
{"root",		'r', 			0, 0,     "Draw on the root window"},
{"fullscreen",	OPT_FULLSCREEN, 0, 0,     "Full screen mode"},
{"fps",			OPT_FPS, 		"NUM", 0, "Frames per second (default = 100 fps)"},
{"speed",		's', 			"NUM", 0, "Speed [10-100] (default = 25)"},
{"gridsize",	'g', 			"NUM", 0, "Gridsize [50-256] (default = 255)"},
{"panoramic",	'p', 			0, 0,     "Panoramic screen mode"},
{0, 0, 0, 0}
};

int parse_opt(int key, char *arg, struct argp_state *state)
{
    int speed, gridsize;

    switch (key) {
	case 'W':
        use_glx = true;
	    window_id = strtol(arg, 0, 0);
	    break;
	case 'r':
        use_glx = true;
	    break;
	case OPT_FULLSCREEN:
	    fullscreen = true;
	    break;
	case OPT_FPS:
	    fps = atoi(arg);
	    break;
	case 's':
        speed = atoi (arg);
        if (speed < 10)
            speed = 10;
        if (speed > 99)
            speed = 99;
        scene.speed = speed;
        break;
	case 'g':
        gridsize = atoi (arg);
        if (gridsize < 50)
            gridsize = 50;
        else if (gridsize > 255)
            gridsize = 255;
        scene.gridsize = gridsize;
        break;
	case 'p':
        panoramic = true;
	    break;
	default:
	    return ARGP_ERR_UNKNOWN;
	    break;
    }
    return 0;
}

static struct argp argp_s = { options, parse_opt, 0, doc };

int main(int argc, char **argv)
{
    if (argp_parse (&argp_s, argc, argv, ARGP_LONG_ONLY, 0, 0) != 0)
        return 0;

    if (use_glx)
        canvas = new CanvasGLX(&scene, fullscreen, fps, window_id);
    else
        canvas = new CanvasSDL(&scene, fullscreen, fps, "Hills", "hills");

    if (canvas->init() < 0) {
        std::cerr << "Can't init display." << std::endl;
        return 0;
    }

    canvas->panoramic = panoramic;
    scene.init ();

    return canvas->loop();
}
