/* 
 * Copyright (c) 2005-2006 Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <cmath>
#include <SDL/SDL_image.h>
#include "texture.h"



Texture::Texture (std::string filename_, bool alpha_, bool repeat_)
{
    filename = filename_;
    alpha    = alpha_;
    repeat   = repeat_;
    id       = 0;
    generate ();
}

Texture::~Texture (void)
{}

void
Texture::generate (void)
{
	GLubyte *data = 0;
	int width, height;

	if ((data = load (width, height)) != 0) {

		int type = (alpha) ? GL_RGBA : GL_RGB;
		glGenTextures (1, &id);
		glBindTexture (GL_TEXTURE_2D, id); 

		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		//  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
		glTexImage2D (GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, data);
		int filter_min, filter_mag;
		filter_min = GL_LINEAR_MIPMAP_LINEAR;
		filter_mag = GL_LINEAR_MIPMAP_LINEAR;

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter_min);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter_mag);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, (repeat) ? GL_REPEAT : GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, (repeat) ? GL_REPEAT : GL_CLAMP);
        gluBuild2DMipmaps(GL_TEXTURE_2D, type, width, height, type, GL_UNSIGNED_BYTE, data);
		delete [] data;
	}
}

GLubyte *
Texture::load (int &width, int &height)
{
	SDL_Surface *img = IMG_Load (filename.c_str());
	if(!img) {
		fprintf(stderr, "Unable to load texture from %s!\n%s\n", filename.c_str(), IMG_GetError());
		width = 0;
		height = 0;
		return 0;
	}
	width = img->w;
	height = img->h;

	// Build image data from the surface (RGB == 3 bytes, RGBA == 4 bytes)
	Uint32 dim = img->w * img->h * ((alpha) ? 4 : 3);
	GLubyte *data = new GLubyte[dim];
	if(!data) {
		fprintf(stderr, "Unable to create a texture from %s\n", filename.c_str());
		width = 0;
		height = 0;
		return 0;
	}

	// Traverse trough surface and grab the pixels
	int pos = 0;
	for (int y=(img->h-1); y>-1; y--) {
		for (int x=0; x<img->w; x++) {
			Uint8 r, g, b, a;
			Uint32 color = pixel (img, x, y);
			if (!alpha)
				SDL_GetRGB (color, img->format, &r, &g, &b);
			else
				SDL_GetRGBA (color, img->format, &r, &g, &b, &a);
			data[pos] = r; pos++;
			data[pos] = g; pos++;
			data[pos] = b; pos++;
			if (alpha) {
				data[pos] = a; pos++;
			}
		}
	}

	SDL_FreeSurface (img);
	return data;
}

Uint32
Texture::pixel (SDL_Surface *surface, int x, int y)
{
	int bpp = surface->format->BytesPerPixel;
	Uint8 *p = (Uint8 *) surface->pixels + y*surface->pitch + x*bpp;

	switch (bpp) {
	case 1:
		return *p;
	case 2:
		return *(Uint16*)p;
	case 3:
		if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
			return p[0] << 16 | p[1] << 8 | p[2];
		else
			return p[0] | p[1] << 8 | p[2] << 16;
	case 4:
		return *(Uint32*)p;
	default:
		return 0;
	}
}



